﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.Knapsack;
using HeuristicLab.Problems.TestFunctions;

namespace HeuristicLab.Algorithms.ScatterSearch.TestFunctions {
  /// <summary>
  /// An operator that improves test functions solutions.
  /// </summary>
  [Item("TestFunctionsImprovementOperator", "An operator that improves test functions solutions.")]
  [StorableClass]
  public abstract class TestFunctionsImprovementOperator : SingleSuccessorOperator, ILocalImprovementOperator, IScatterSearchTargetProcessor {
    #region Problem properties
    public Type ProblemType {
      get { return typeof(KnapsackProblem); }
    }
    [Storable]
    private KnapsackProblem problem;
    public IProblem Problem {
      get { return problem; }
      set { problem = (KnapsackProblem)value; }
    }
    #endregion

    #region Parameter properties
    public ScopeParameter CurrentScopeParameter {
      get { return (ScopeParameter)Parameters["CurrentScope"]; }
    }
    public IValueLookupParameter<DoubleMatrix> CoordinatesParameter {
      get { return (IValueLookupParameter<DoubleMatrix>)Parameters["Coordinates"]; }
    }
    public IValueLookupParameter<ISingleObjectiveTestFunctionProblemEvaluator> EvaluatorParameter {
      get { return (IValueLookupParameter<ISingleObjectiveTestFunctionProblemEvaluator>)Parameters["Evaluator"]; }
    }
    public IValueLookupParameter<IntValue> ImprovementAttemptsParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["ImprovementAttempts"]; }
    }
    public IValueLookupParameter<IRandom> RandomParameter {
      get { return (IValueLookupParameter<IRandom>)Parameters["Random"]; }
    }
    public IValueLookupParameter<IItem> TargetParameter {
      get { return (IValueLookupParameter<IItem>)Parameters["Target"]; }
    }
    public IValueParameter<DoubleValue> AlphaParameter {
      get { return (IValueParameter<DoubleValue>)Parameters["Alpha"]; }
    }
    public IValueParameter<DoubleValue> BetaParameter {
      get { return (IValueParameter<DoubleValue>)Parameters["Beta"]; }
    }
    public IValueParameter<DoubleValue> GammaParameter {
      get { return (IValueParameter<DoubleValue>)Parameters["Gamma"]; }
    }
    public IValueParameter<DoubleValue> DeltaParameter {
      get { return (IValueParameter<DoubleValue>)Parameters["Delta"]; }
    }
    #region ILocalImprovementOperator Parameters
    public IValueLookupParameter<IntValue> MaximumIterationsParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["MaximumIterations"]; }
    }
    public ILookupParameter<IntValue> EvaluatedSolutionsParameter {
      get { return (ILookupParameter<IntValue>)Parameters["EvaluatedSolutions"]; }
    }
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    #endregion
    #endregion

    #region Properties
    private IItem Target {
      get { return TargetParameter.ActualValue; }
    }
    private DoubleValue Alpha {
      get { return AlphaParameter.Value; }
    }
    private DoubleValue Beta {
      get { return BetaParameter.Value; }
    }
    private DoubleValue Gamma {
      get { return GammaParameter.Value; }
    }
    private DoubleValue Delta {
      get { return DeltaParameter.Value; }
    }
    protected Func<RealVector, double> FunctionEvaluator { get; set; }
    public IScope CurrentScope {
      get { return CurrentScopeParameter.ActualValue; }
    }
    public DoubleMatrix Coordinates {
      get { return CoordinatesParameter.ActualValue; }
      set { CoordinatesParameter.ActualValue = value; }
    }
    public ISingleObjectiveTestFunctionProblemEvaluator Evaluator {
      get { return EvaluatorParameter.ActualValue; }
      set { EvaluatorParameter.ActualValue = value; }
    }
    public IntValue ImprovementAttempts {
      get { return ImprovementAttemptsParameter.ActualValue; }
      set { ImprovementAttemptsParameter.ActualValue = value; }
    }
    public IRandom Random {
      get { return RandomParameter.ActualValue; }
      set { RandomParameter.ActualValue = value; }
    }
    #endregion

    [StorableConstructor]
    protected TestFunctionsImprovementOperator(bool deserializing) : base(deserializing) { }
    protected TestFunctionsImprovementOperator(TestFunctionsImprovementOperator original, Cloner cloner)
      : base(original, cloner) {
      this.problem = cloner.Clone(original.problem);
    }
    public TestFunctionsImprovementOperator()
      : base() {
      #region Create parameters
      Parameters.Add(new ScopeParameter("CurrentScope"));
      Parameters.Add(new ValueLookupParameter<DoubleMatrix>("Coordinates"));
      Parameters.Add(new ValueLookupParameter<ISingleObjectiveTestFunctionProblemEvaluator>("Evaluator"));
      Parameters.Add(new ValueLookupParameter<IntValue>("ImprovementAttempts", new IntValue(100)));
      Parameters.Add(new ValueLookupParameter<IRandom>("Random"));
      Parameters.Add(new ValueLookupParameter<IItem>("Target"));
      Parameters.Add(new ValueParameter<DoubleValue>("Alpha", new DoubleValue(1.0)));
      Parameters.Add(new ValueParameter<DoubleValue>("Beta", new DoubleValue(2.0)));
      Parameters.Add(new ValueParameter<DoubleValue>("Gamma", new DoubleValue(0.5)));
      Parameters.Add(new ValueParameter<DoubleValue>("Delta", new DoubleValue(0.5)));
      #endregion
      TargetParameter.ActualName = "Point"; // temporary solution for the test functions problem
    }

    public override IOperation Apply() {
      var bestSol = CurrentScope.Variables[TargetParameter.ActualName].Value as RealVector;
      var bestSolQuality = FunctionEvaluator(bestSol);

      // create perturbed solutions
      var simplex = new RealVector[bestSol.Length];
      for (int i = 0; i < simplex.Length; i++) {
        simplex[i] = bestSol.Clone() as RealVector;
        simplex[i][i] += 0.1 * (Evaluator.Bounds[0, 1] - Evaluator.Bounds[0, 0]);
        if (simplex[i][i] > Evaluator.Bounds[0, 1]) simplex[i][i] = Evaluator.Bounds[0, 1];
        if (simplex[i][i] < Evaluator.Bounds[0, 0]) simplex[i][i] = Evaluator.Bounds[0, 0];
      }

      // improve solutions
      for (int i = 0; i < ImprovementAttempts.Value; i++) {
        // order according to their objective function value
        Array.Sort(simplex, (x, y) => FunctionEvaluator(x).CompareTo(FunctionEvaluator(y)));

        // calculate centroid
        var centroid = new RealVector(bestSol.Length);
        foreach (var vector in simplex)
          for (int j = 0; j < centroid.Length; j++)
            centroid[j] += vector[j];
        for (int j = 0; j < centroid.Length; j++)
          centroid[j] /= simplex.Length;

        // reflection
        var reflectionPoint = new RealVector(bestSol.Length);
        for (int j = 0; j < reflectionPoint.Length; j++)
          reflectionPoint[j] = centroid[j] + Alpha.Value * (centroid[j] - simplex[simplex.Length - 1][j]);
        double reflectionPointQuality = FunctionEvaluator(reflectionPoint);
        if (FunctionEvaluator(simplex[0]) <= reflectionPointQuality
            && reflectionPointQuality < FunctionEvaluator(simplex[simplex.Length - 2]))
          simplex[simplex.Length - 1] = reflectionPoint;

        // expansion
        if (reflectionPointQuality < FunctionEvaluator(simplex[0])) {
          var expansionPoint = new RealVector(bestSol.Length);
          for (int j = 0; j < expansionPoint.Length; j++)
            expansionPoint[j] = centroid[j] + Beta.Value * (reflectionPoint[j] - centroid[j]);
          simplex[simplex.Length - 1] = FunctionEvaluator(expansionPoint) < reflectionPointQuality ? expansionPoint : reflectionPoint;
        }

        // contraction
        if (FunctionEvaluator(simplex[simplex.Length - 2]) <= reflectionPointQuality
            && reflectionPointQuality < FunctionEvaluator(simplex[simplex.Length - 1])) {
          var outsideContractionPoint = new RealVector(bestSol.Length);
          for (int j = 0; j < outsideContractionPoint.Length; j++)
            outsideContractionPoint[j] = centroid[j] + Gamma.Value * (reflectionPoint[j] - centroid[j]);
          if (FunctionEvaluator(outsideContractionPoint) <= reflectionPointQuality) {
            simplex[simplex.Length - 1] = outsideContractionPoint;
            if (FunctionEvaluator(reflectionPoint) >= FunctionEvaluator(simplex[simplex.Length - 1])) {
              var insideContractionPoint = new RealVector(bestSol.Length);
              for (int j = 0; j < insideContractionPoint.Length; j++)
                insideContractionPoint[j] = centroid[j] - Gamma.Value * (reflectionPoint[j] - centroid[j]);
              if (FunctionEvaluator(insideContractionPoint) < FunctionEvaluator(simplex[simplex.Length - 1])) simplex[simplex.Length - 1] = insideContractionPoint;
            }
          }
        }

        // reduction
        for (int j = 1; j < simplex.Length; j++)
          for (int k = 0; k < simplex[j].Length; k++)
            simplex[j][k] = simplex[0][k] + Delta.Value * (simplex[j][k] - simplex[0][k]);
      }

      for (int i = 0; i < simplex[0].Length; i++) {
        if (simplex[0][i] > Evaluator.Bounds[0, 1]) simplex[0][i] = Evaluator.Bounds[0, 1];
        if (simplex[0][i] < Evaluator.Bounds[0, 0]) simplex[0][i] = Evaluator.Bounds[0, 0];
      }

      CurrentScope.Variables[TargetParameter.ActualName].Value = simplex[0];

      return base.Apply();
    }
  }
}
